/****************************************************************************
   Copyright & License:
   ====================
   
   Copyright 2009 - 2015 gAGE/UPC & ESA
   
   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 ***************************************************************************/

/****************************************************************************
 * Copyright: gAGE/UPC & ESA
 * Project: EDUNAV GNSS Lab Tool 
 * Author: Pere Ramos-Bosch (group of Astronomy and GEomatics - gAGE/UPC)
 * Curator: Adria Rovira-Garcia ( gAGE/UPC )
 * Developer: Deimos Ibanez Segura ( gAGE/UPC )     
 *          glab.gage @ upc.edu
 * File: filter.c
 * Code Management Tool File Version: 2.2  Revision: 3
 * Date: 2015/05/28
 ***************************************************************************/

/****************************************************************************
 * MODULE DESCRIPTION
 *
 * Name: filter
 * Language: C
 *
 * Purpose:
 *  The purpose of this module is to contain all the necessary functions to
 *  do a Kalman filtering process.
 *
 * Function:
 *  This class is in charge of declaring providing all the necessary functions
 *  to do a Kalman filtering process. It has a high level function (Kalman) in
 *  which, given all necessary inputs, provides the estimated values for the
 *  parameters 
 *
 * Dependencies:
 *  dataHandling, model
 *
 * Files modified:
 *  None
 *
 * Files read:
 *  None
 *
 * Resources usage:
 *  See Design Document
 ****************************************************************************/

/****************************
 *       RELEASE_HISTORY
 * -------------
 *          gLAB v1.3.1
 * Release: 2010/01/22
 * Change Log: First public version
 * -------------
 *          gLAB v1.3.2
 * Release: 2010/03/23
 * Change Log: PREFIT message now prints the weight to be applied in the filter
 *             FILTER message now prints the total ZTD and the estimated ZTD (substracting the modeled value)
 *             PREFIT message now prints non-GPS satellites
 * -------------
 *          gLAB v1.4.0
 * Release: 2010/06/21
 * Change Log: Included usable measurement index to select/deselect individual measurements
 * -------------
 *          gLAB v1.4.1
 * Release: 2010/07/09
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.2
 * Release: 2010/07/31
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.3
 * Release: 2010/08/31
 * Change Log: Improved DCB Management.
 * -------------
 *          gLAB v1.4.4
 * Release: 2010/09/22
 * Change Log: Filter adjustments to new DCB management.
 * -------------
 *          gLAB v1.4.5
 * Release: 2010/10/31
 * Change Log: Dilution Of Precision extension: (GDOP, PDOP, TDOP, VDOP, HDOP).
 * -----------
 *          gLAB v2.0.0
 * Release: 2012/12/31
 * Change Log: gLAB version released together with the Educational Book:
 *             "GNSS Data Processing" (Vol. 1 and Vol. 2). ESA TM-23.
 *             Authors: J. Sanz Subirana, J.M. Juan Zornoza and M. Hernandez-Pajares
 *             ISBN: 978-92-9221-885-0 (two volumes)
 *             ISSN: 1013-7076
 *             December 2012.
 * -----------
 *          gLAB v2.2.0
 * Release: 2014/09/22
 * Change Log: Modified fillUsableSatellites function in order to discard satellites without IONEX or F-PPP model data
 *                      (in case those models are used).
 * -----------
 *          gLAB v2.2.1
 * Release: 2014/12/23
 * Change Log: Fixed initialization to given value per parameter of covariance of coordinates Y and Z (before only coordinate X was initialized, Y and Z were 0)
 * -----------
 *          gLAB v2.2.2
 * Release: 2015/03/02
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.3
 * Release: 2015/05/28 
 * Change Log: No changes in this file.
 * -----------
 *       END_RELEASE_HISTORY
 *****************************/

/* External classes */
#include "filter.h"

/**************************************
 * Declarations of internal operations
 **************************************/

/*****************************************************************************
 * Name        : m2v
 * Description : Transforms the indices from matrix to vector
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int  i                          I  N/A  Matrix row
 * int  j                          I  N/A  Matrix column
 * Returned value (int)            O  N/A  Vector index
 *****************************************************************************/
int m2v (int i, int j) {
  return i>j ? i*(i+1)/2+j : j*(j+1)/2+i;
}

/*****************************************************************************
 * Name        : m2v_opt
 * Description : Optimized version of m2v (much faster). It suposes i>j.
 *               It is specially useful when inverting matrixes.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int  i                          I  N/A  Matrix row
 * int  j                          I  N/A  Matrix column
 * Returned value (int)            O  N/A  Vector index
 *****************************************************************************/
static inline int m2v_opt (int i, int j) {
  return i*(i+1)/2+j;
}

/*****************************************************************************
 * Name        : cholinv
 * Description : Cholesky matrix inversion
 *               Version inspired from "Numerical Recipes in C" P.97-98
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *mat                    IO N/A  Matrix in vector form (overwritten)
 * int  n                          I  N/A  Matrix size
 * Returned value (int)            O  N/A  Status of the function
 *                                         0  => Cholesky worked properly
 *                                         -1 => Cholesky inversion failed (singular
 *                                               matrix)
 *****************************************************************************/
int cholinv(double *mat, int n) {
  int i,j,k;
  double sum;

  for (i=0;i<n;i++) {
    for (j=i;j<n;j++) {
      for (sum=mat[m2v(i,j)],k=i-1;k>=0;k--) {sum -= mat[m2v(i,k)]*mat[m2v(j,k)]; } 
      if (i==j) {
        if (sum <= 0) return -1;  // Cholesky decomposition failed
        mat[m2v(i,i)]=sqrt(sum);
      } else mat[m2v(j,i)]=sum/mat[m2v(i,i)];
    }
  }

// Beginning Upper Inversion

  for (i=0;i<n;i++) {
    mat[m2v(i,i)]=1/mat[m2v(i,i)];
    for (j=i+1;j<n;j++) {
      sum=0;
      for (k=i;k<j;k++) sum -= mat[m2v(j,k)]*mat[m2v(k,i)];
      mat[m2v(j,i)]=sum/mat[m2v(j,j)];
    }
  }

// Multiplying U^-1 * UT^-1

  for (i=0;i<n;i++) {
    for (j=i;j<n;j++) {
      sum=0;
      for (k=j;k<n;k++) sum += mat[m2v(i,k)]*mat[m2v(k,j)];
      mat[m2v(i,j)]=sum;
    }
  }

  return 0;
}

/*****************************************************************************
 * Name        : cholinv_opt
 * Description : Optimized version of cholinv. It uses m2v_opt
 *               Version inspired from "Numerical Recipes in C" P.97-98
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *mat                    IO N/A  Matrix in vector form (overwritten)
 * int  n                          I  N/A  Matrix size
 * Returned value (int)            O  N/A  Status of the function
 *                                         0  => Cholesky worked properly
 *                                         -1 => Cholesky inversion failed (singular
 *                                               matrix)
 *****************************************************************************/
int cholinv_opt(double *mat, int n) {
  int i,j,k;
  double sum;
  double *aux;
 
  // Reserving space 
  aux = malloc((m2v_opt(n-1,n-1)+1)*sizeof(double));
  memset(aux,0,(m2v_opt(n-1,n-1)+1)*sizeof(double));

  for (i=0;i<n;i++) {
    for (j=i;j<n;j++) {
      for (sum=mat[m2v_opt(j,i)],k=i-1;k>=0;k--) sum -= mat[m2v_opt(i,k)]*mat[m2v_opt(j,k)];
      if (i==j) {
        if (sum <= 0) {
        	free(aux);
        	return -1;  // Cholesky decomposition failed
        }
  		mat[m2v_opt(i,i)]=sqrt(sum);
      } else mat[m2v_opt(j,i)]=sum/mat[m2v_opt(i,i)];
    }
  }

// Upper Inversion and multiplication of U^-1 * UT^-1

  for (i=0;i<n;i++) {
    mat[m2v_opt(i,i)]=1/mat[m2v_opt(i,i)];
    for (k=0;k<=i;k++) aux[m2v_opt(i,k)]+=mat[m2v_opt(i,i)]*mat[m2v_opt(i,k)];
    for (j=i+1;j<n;j++) {
      sum=0;
      for (k=i;k<j;k++) sum -= mat[m2v_opt(j,k)]*mat[m2v_opt(k,i)];
      mat[m2v_opt(j,i)]=sum/mat[m2v_opt(j,j)];
      for (k=0;k<=i;k++) aux[m2v_opt(i,k)]+=mat[m2v_opt(j,i)]*mat[m2v_opt(j,k)];
    }
  }

  memcpy(mat,aux,(m2v_opt(n-1,n-1)+1)*sizeof(double));

  free (aux);
  return 0;
}

/*****************************************************************************
 * Name        : mxv
 * Description : Multiplies a matrix in vectorial form and a vector
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *mat                    I  N/A  Matrix in vector form
 * double  *vec                    IO N/A  Vector (overwritten)
 * int  n                          I  N/A  Matrix and vector size
 *****************************************************************************/
void mxv(double *mat, double *vec, int n) {
	int		i,j;
	double	*aux;

	aux=(double *)(malloc(sizeof(double)*n));

	for (i=0;i<n;i++) {
		aux[i]=0;
		for (j=0;j<n;j++) {
			aux[i] += mat[m2v(i,j)] * vec[j];
		}
	}
	for (i=0;i<n;i++) vec[i]=aux[i];
	free(aux);
}

/*****************************************************************************
 * Name        : fillUsableSatellites
 * Description : Check for the usable satellites in the specified epoch
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  Reference epoch
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void fillUsableSatellites(TEpoch *epoch, TOptions *options) {
	int		i,j;
	int		check;
	double	sunPos[3];
	double	auxT;
  int     onlyHasPCorPCC;
  enum MeasurementType   meas;

	epoch->usableSatellites = 0;
	
	if (options->discardEclipsedSatellites) {
		findSun(&epoch->t,sunPos);
	}

	for (i=0;i<epoch->numSatellites;i++) {
		check = 1;

		// Arc length sanity check
		if (epoch->sat[i].arcLength < options->minArcLength) {
			sprintf(messagestr,"Arc too short (%d)",epoch->sat[i].arcLength);
			printSatSel(0,epoch->sat[i].PRN,messagestr,options);
			check = 0;
		}

		// Data availability check
		for (j=0;j<options->totalFilterMeasurements;j++) {
			// Check if there are valid measurements for navigating
			if (getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,options->measurement[j],NULL,NULL)!=1) {
				sprintf(messagestr,"Measurement unavailable (%s)",meastype2measstr(options->measurement[j]));
				printSatSel(0,epoch->sat[i].PRN,messagestr,options);
				check = 0; 
			}
		}
		
		// Satellite selection check
		if (!options->includeSatellite[epoch->sat[i].GNSS][epoch->sat[i].PRN]) {
			printSatSel(0,epoch->sat[i].PRN,"Deselection forced",options);
			check = 0;
		}

		// Ephemeris availability check
		if (epoch->sat[i].hasOrbitsAndClocks == 0) {
			printSatSel(0,epoch->sat[i].PRN,"Ephemerides unavailable",options);
			check = 0;
		}
		
		// DCBs availability check
		if (epoch->sat[i].hasDCBs == 0) {
		      if (epoch->sat[i].hasDCBsofC1P1 == 1) {
			onlyHasPCorPCC = 1;
			for (j=0;j<options->totalFilterMeasurements;j++) {
			  meas = options->measurement[j];
			  if (whatIs(meas)==Pseudorange && meas!=PC && meas!=PCC) onlyHasPCorPCC = 0;
			}
		      }
		      if (!(epoch->sat[i].hasDCBs==1 || (epoch->sat[i].hasDCBsofC1P1==1 && onlyHasPCorPCC==1))) {
					  printSatSel(0,epoch->sat[i].PRN,"DCBs unavailable (strict TGD treatment)",options);
					  check = 0;
		      }
		}
		
		// Elevation check (if approximate position is good enough, ie <100Km)
		if ((epoch->sat[i].elevation < options->elevationMask) && (epoch->receiver.aproxPositionError < 1e5)) { // 100Km
			sprintf(messagestr,"Elevation too low (%7.2lf)",epoch->sat[i].elevation/d2r);
			printSatSel(0,epoch->sat[i].PRN,messagestr,options);
			check = 0;
		}
		
		// Eclipsed satellite check
		if (options->discardEclipsedSatellites) {
			if (isEclipsed(&epoch->t,&epoch->sat[i],&epoch->satInfo[epoch->sat[i].GNSS][epoch->sat[i].PRN],sunPos,&auxT)) {
				if (auxT==0) {
					printSatSel(0,epoch->sat[i].PRN,"Satellite under eclipse",options);
				} else {
					sprintf(messagestr,"Satellite was under eclipse %4.0lf seconds ago",auxT);
					printSatSel(0,epoch->sat[i].PRN,messagestr,options);
				}
				check = 0;
			}
		}

		// Ionex check (in case a IONEX file is used)
		if (options->ionoModel == IONEX && epoch->sat[i].validIonex == 0) {
			sprintf(messagestr,"Missing IONEX data");
                        printSatSel(0,epoch->sat[i].PRN,messagestr,options);
                        check = 0;
                }

		// FPPP check (in case a FPPP file is used)
                if (options->ionoModel == FPPP && epoch->sat[i].validIonex == 0) {
                        sprintf(messagestr,"Missing FPPP data");
                        printSatSel(0,epoch->sat[i].PRN,messagestr,options);
                        check = 0;
                }



		// GNSS System check (only prepared for GPS processing)
		if (epoch->sat[i].GNSS != GPS) {
			sprintf(messagestr,"Invalid GNSS System: %s",gnsstype2gnssstr(epoch->sat[i].GNSS));
			printSatSel(0,epoch->sat[i].PRN,messagestr,options);
			check = 0;
		}
		

		// If everything is available satellite can be used
		if (check) {
			epoch->sat[i].available = 1;
			printSatSel(1,epoch->sat[i].PRN,"",options);
			epoch->usableSatellites++;
		} else {
			epoch->sat[i].available = 0;
		}
	}
}

/*****************************************************************************
 * Name        : isEpochComputable
 * Description : Check (taking into account the usable satellites) if the 
 *               specified epoch can be computed
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  Reference epoch
 * Returned value (int)            O  N/A  Status of the function
 *                                         0 => Epoch not computable
 *                                         1 => Epoch computable
 *****************************************************************************/
int isEpochComputable(TEpoch *epoch) {
	if (epoch->usableSatellites>=4) return 1;
	else return 0;
}

/*****************************************************************************
 * Name        : calculateUnknowns
 * Description : Computes, from the options specified, how many unknowns the 
 *               filter has.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int  obs                        I  N/A  Usable observations
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (int)            O  N/A  Number of unknowns in the filter
 *****************************************************************************/
int calculateUnknowns(int obs, TOptions *options) {
	int	value = 4;

	if (options->usePhase==1) value+=obs;
	if (options->estimateTroposphere==1) value++;
	return value;
}

/*****************************************************************************
 * Name        : initUnkinfo
 * Description : Initialises the information for the unknwons of the filter
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TUnkinfo  *unkinfo              O  N/A  TUnkinfo structure to store unknows
 * int  nunk                       I  N/A  Number of unknowns
 * int  observations               I  N/A  Number of observations
 * double  t                       I  s    Time since last filter run
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void initUnkinfo(TEpoch *epoch, TUnkinfo *unkinfo, int nunk, int observations, double t, TOptions *options){
	int	i;

	unkinfo->nunk=nunk;
	unkinfo->nunkvector=m2v(nunk-1,nunk-1)+1;
	unkinfo->par2unk[DR_UNK]=0;
	unkinfo->par2unk[DT_UNK]=3;
	i=4;
	if (options->estimateTroposphere==1) {
		unkinfo->par2unk[TROP_UNK]=i;
		i++;
	}
	if (options->usePhase==1) {
		unkinfo->par2unk[BIAS_UNK]=i;
		i+=observations;
	}
	
	// Forming PHI and QNOISE MATRIX (ONLY DIAGONAL)
	for (i=0;i<3;i++) {
		unkinfo->phi[unkinfo->par2unk[DR_UNK]+i] = options->filterParams[PHI_PAR][DR_UNK];
		unkinfo->qnoise[unkinfo->par2unk[DR_UNK]+i] = options->filterParams[Q_PAR][DR_UNK];
	}
	unkinfo->phi[unkinfo->par2unk[DT_UNK]] = options->filterParams[PHI_PAR][DT_UNK];
	unkinfo->qnoise[unkinfo->par2unk[DT_UNK]] = options->filterParams[Q_PAR][DT_UNK];

	if (options->estimateTroposphere==1) {
		unkinfo->phi[unkinfo->par2unk[TROP_UNK]] = options->filterParams[PHI_PAR][TROP_UNK];
		unkinfo->qnoise[unkinfo->par2unk[TROP_UNK]] = options->filterParams[Q_PAR][TROP_UNK]*(t/3600.);
	}
	if (options->usePhase==1) {
		for (i=0;i<observations;i++) {
			unkinfo->phi[unkinfo->par2unk[BIAS_UNK]+i] = options->filterParams[PHI_PAR][BIAS_UNK];
			unkinfo->qnoise[unkinfo->par2unk[BIAS_UNK]+i] = options->filterParams[Q_PAR][BIAS_UNK];
		}
	}
}

/*****************************************************************************
 * Name        : prepareCorrelation
 * Description : Updates the correlations for the solution computation
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TFilterSolution  *solution      IO N/A  Structure containing the filter 
 *                                         information of the previous epoch
 * int  *PRNlist                   I  N/A  List of satellites of the previous epoch
 * TUnkinfo  *unkinfo              I  N/A  TUnkinfo structure with information on 
 *                                         the current epoch unknowns
 * TUnkinfo  *prevUnkinfo          I  N/A  TUnkinfo structure with information on 
 *                                         the previous epoch unknowns
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void prepareCorrelation(TEpoch *epoch, TFilterSolution *solution, int *PRNlist, TUnkinfo *unkinfo, TUnkinfo *prevUnkinfo, TOptions *options) {
	int		i,j,alti,altj;
	double	updatedCorrelations[MAX_VECTOR_UNK];
 	double	updatedParameterValues[MAX_UNK];
 	int		new2old[MAX_SATELLITES_IN_VIEW];
	int		iniBiasUnk,endBiasUnk;
	int		new;
	int		found;
	int		ind;
	int		res;
	
	// Check for FIRST time
	if (solution->prevNumSatellites==0) {
		if (options->estimateTroposphere==1) {
			solution->correlations[m2v(unkinfo->par2unk[TROP_UNK],unkinfo->par2unk[TROP_UNK])] = 1/(options->filterParams[P0_PAR][TROP_UNK]);
		}
                //Loop to initialize covariance for all coordinates
                for (i=0;i<3;i++) {
                        solution->correlations[m2v(unkinfo->par2unk[DR_UNK]+i,unkinfo->par2unk[DR_UNK]+i)] = 1/(options->filterParams[P0_PAR][DR_UNK]);
                }
		solution->correlations[m2v(unkinfo->par2unk[DT_UNK],unkinfo->par2unk[DT_UNK])] = 1/(options->filterParams[P0_PAR][DT_UNK]);
	}

	// Vector new2old will have the dictionary from new PRN positions to previous PRN positions
	for (i=0;i<epoch->numSatellites;i++) {
		new2old[i]=-1;
		if (epoch->sat[i].GNSS==GPS) {
			for (j=0;j<solution->prevNumSatellites;j++) {
				if (PRNlist[j]==epoch->sat[i].PRN && epoch->sat[i].available && !epoch->sat[i].hasBeenCycleslip) {
					new2old[i]=j;
     			}
			}
		}
		if (epoch->sat[i].hasBeenCycleslip) epoch->sat[i].hasBeenCycleslip = 0; // clear cycle slip flag
	}
	if (options->usePhase==1) {
		iniBiasUnk=unkinfo->par2unk[BIAS_UNK];
		endBiasUnk=unkinfo->par2unk[BIAS_UNK]+epoch->numSatellites-1;
	} else {
		iniBiasUnk=1;
		endBiasUnk=0;
	}
	
	// Updating correlations (including Phi and Q)
	if (solution->prevNumSatellites!=0) {
		for (i=0;i<prevUnkinfo->nunk;i++) {
			for (j=i;j<prevUnkinfo->nunk;j++) {
				solution->correlations[m2v(i,j)] *= prevUnkinfo->phi[i]*prevUnkinfo->phi[j];
			}
			solution->correlations[m2v(i,i)] += prevUnkinfo->qnoise[i];
		}
	}
	
	// Updating parameters
	for (i=0;i<prevUnkinfo->nunk;i++) {
		solution->parameters[i] *= prevUnkinfo->phi[i];
	}
	
	if (solution->prevNumSatellites!=0) { // If not in the first epoch
		// Update position parameters
		for(i=0,j=unkinfo->par2unk[DR_UNK];i<3;i++,j++) {
			solution->parameters[j] = solution->x[j] - epoch->receiver.aproxPosition[i];
		}
		// Update tropospheric parameter
		if (options->estimateTroposphere) {
			j = unkinfo->par2unk[TROP_UNK];
			solution->parameters[j] =  solution->x[j] - epoch->ZTD;
		}
	}
	
	// Satellite ambiguities that have disappeared or that have had a cycle-slip will be decorrelated from the
	// rest of ambiguities by putting an infinite in its self-correlation
	if (solution->prevNumSatellites!=0 && options->usePhase) {
		for (j=0;j<solution->prevNumSatellites;j++) {
			found = 0;
			for (i=0;i<epoch->numSatellites;i++) {
				if (new2old[i]==j) {
					found = 1;
					break;
				}
			}
			if (!found) {
				ind = prevUnkinfo->par2unk[BIAS_UNK]+j;
				solution->correlations[m2v(ind,ind)] = (options->filterParams[P0_PAR][BIAS_UNK]);
			}
		}
	}
	
	// Inverting correlation matrix from the former epoch
	res = cholinv_opt(solution->correlations,prevUnkinfo->nunk);
	if (res==-1) {
		sprintf(messagestr,"Problem in CHOLESKI2 res=%d\n",res);
		printError(messagestr,options);
		return;
	}
	
	// Creating new correlation matrix
	for (i=0;i<unkinfo->nunk;i++) {
		if (i>=iniBiasUnk && i<=endBiasUnk) {
			if (new2old[i-iniBiasUnk]!=-1) {
				updatedParameterValues[i] = solution->parameters[new2old[i-iniBiasUnk]+iniBiasUnk];
			} else {
				updatedParameterValues[i] = 0;
			}
		} else {
			updatedParameterValues[i] = solution->parameters[i];
		}
		for (j=i;j<unkinfo->nunk;j++) {
			new=0;
			if ((i>=iniBiasUnk && i<=endBiasUnk) || (j>=iniBiasUnk && j<=endBiasUnk)) {
				if (i>=iniBiasUnk && i<=endBiasUnk) {
					if (new2old[i-iniBiasUnk]==-1) {new=1;
					} else { alti = prevUnkinfo->par2unk[BIAS_UNK]+new2old[i-iniBiasUnk]; }
				} else { alti = i;}
				if (j>=iniBiasUnk && j<=endBiasUnk) {
					if (new2old[j-iniBiasUnk]==-1) { new=1;
					} else { altj = prevUnkinfo->par2unk[BIAS_UNK]+new2old[j-iniBiasUnk];}
				} else { altj = j;}
  				if (new) {
					if (i==j) updatedCorrelations[m2v(i,j)] = 1/(options->filterParams[P0_PAR][BIAS_UNK]);
					else updatedCorrelations[m2v(i,j)] = 0;
				} else {
					updatedCorrelations[m2v(i,j)] = solution->correlations[m2v(alti,altj)];
				}
			} else {
				updatedCorrelations[m2v(i,j)] = solution->correlations[m2v(i,j)];
			}
		}
	}

	memcpy(solution->correlations,updatedCorrelations,sizeof(double)*unkinfo->nunkvector);
	memcpy(solution->parameters,updatedParameterValues,sizeof(double)*unkinfo->nunk);
}

/*****************************************************************************
 * Name        : atwa_atwy_insertline
 * Description : Insert an equation (prefilter values and design vector) into 
 *               the AtWA matrix and AtWY vector
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *atwa                   IO N/A  At*W*A matrix
 * double  *atwy                   IO N/A  At*W*Y vector
 * double  *unk                    I  N/A  Design vector
 * double  prefit                  I  N/A  Prefit value
 * int  nunk                       I  N/A  Number of unknowns
 * double  sigma2                  I  N/A  Sigma^2 associated with the prefit
 *****************************************************************************/
void atwa_atwy_insertline (double *atwa, double *atwy, double *unk, double prefit, int nunk, double sigma2) {
	int j,k;

	for (j=0;j<nunk;j++) {
		atwy[j] = atwy[j] + unk[j]*prefit/sigma2;
		for (k=0;k<=j;k++) {
			atwa[m2v(k,j)] = atwa[m2v(k,j)] + unk[j]*unk[k]/sigma2;
		}
	}
}

/*****************************************************************************
 * Name        : atwa_atwy_insertLineWithCorrelation
 * Description : Similar to atwa_atwy_insert for apriori values (with a full 
 *               correlations vector)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *atwa                   IO N/A  At*W*A matrix
 * double  *atwy                   IO N/A  At*W*Y vector
 * double  *corr                   I  N/A  Correlations vector
 * double  apriorivalue            I  N/A  Apriori value (equivalent to prefit)
 * int  nunk                       I  N/A  Number of unknowns
 * double  multiplier              I  N/A  Multiplier in the Fi matrix (equivalent
 *                                         to the design vector)
 * int  iniunk                     I  N/A  Initial unk to start updating
 *****************************************************************************/
void atwa_atwy_insertLineWithCorrelation (double *atwa, double *atwy, double *corr, double apriorivalue, int nunk, double multiplier, int iniunk) {
	int i;

	for (i=0;i<nunk;i++) {
		atwy[i] += multiplier*apriorivalue*corr[i];
	}

	for (i=iniunk;i<nunk;i++) {
		atwa[m2v(iniunk,i)] += multiplier*multiplier*corr[i];
	}
}

/*****************************************************************************
 * Name        : designSystem
 * Description : Designs the system for a specific measurement marked in the filter.
 *               Computes At*W*A and At*W*Y for the specified epoch 
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * double  *atwa                   IO N/A  At*W*A matrix
 * double  *atwy                   IO N/A  At*W*Y vector
 * TUnkinfo  *unkinfo              I  N/A  Structure containing the information of
 *                                         the unknowns 
 * MeasurementType  measType	   I  N/A  Measurement to be included in the filter
 * int  filterInd                  I  N/A  Index of the measurement of the filter
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void designSystem(TEpoch *epoch, double *atwa, double *atwy, TUnkinfo *unkinfo, enum MeasurementType measType, int filterInd, TOptions *options) {
	int			i,j;
	double		unk[MAX_UNK],prefit,measurement,model;
	int			ns=0;
	TSatellite	*sat;
	double		weight;
	double		stdDev;
	int			ret;
	int			sats[MAX_SATELLITES_IN_VIEW];
	double		smoothWith;
	int			dontuse;

	for (i=0;i<epoch->numSatellites;i++) {
		dontuse = 0;
		sat = &epoch->sat[i];
		
		// If not GPS, the satellite will not be included in the filter
		if (sat->GNSS != GPS) dontuse=1;

		ret = getMeasModelValue(epoch,sat->GNSS,sat->PRN,measType,&measurement,&model);
		if (ret!=1) {
			sat->available=0;
			model = 0;
			measurement = 0;
		}
		// Smoothing
		if (options->smoothEpochs!=0 && options->smoothMeas[filterInd]!=NA) {
			ret = getMeasModelValue(epoch,sat->GNSS,sat->PRN,options->smoothMeas[filterInd],&smoothWith,NULL);
			if (ret!=1) {
				sat->available=0;
				model = 0;
				smoothWith = 0;
				sat->smoothedMeas[filterInd] = measurement;
			}
			prefit = sat->smoothedMeas[filterInd] + smoothWith - model;
		} else {
			prefit = measurement - model;
		}

		// Compute weights and available satellites
		if (sat->available) {
			stdDev = sat->measurementWeights[filterInd];
			sats[ns] = sat->PRN;
			ns++;
		} else {
			// Satellite not counted for number of satellites
			stdDev = SIGMA_INF;
		}
		weight = stdDev * stdDev;
	
		// Stop if it is declared as Don't Use
		if (dontuse) continue;
		
		// Print prefit residuals
		printPrefit(&epoch->t,sat,measType,prefit,measurement,model,stdDev,options);
		
		// unk vector is the a row of the design matrix
		for (j=0;j<unkinfo->nunk;j++) unk[j]=0;

		// Position unknowns
		unk[unkinfo->par2unk[DR_UNK]] = -sat->LoS[0];
		unk[unkinfo->par2unk[DR_UNK]+1] = -sat->LoS[1];
		unk[unkinfo->par2unk[DR_UNK]+2] = -sat->LoS[2];

		// Time unknown
		unk[unkinfo->par2unk[DT_UNK]] = 1;

		// Troposhpere unknown
		if (options->estimateTroposphere)
			unk[unkinfo->par2unk[TROP_UNK]] = sat->tropWetMap;
			
		// Carrier-phase ambiguities unknwowns - if applicable
		if (whatIs(measType)==CarrierPhase) {
			unk[unkinfo->par2unk[BIAS_UNK]+i] = 1;
		}
		
		// Insert the data in atwa and atwy. Only include if available or it is a carrierphase
		if (sat->available || whatIs(measType)==CarrierPhase) {
			atwa_atwy_insertline(atwa,atwy,unk,prefit,unkinfo->nunk,weight);
		}
		
	}

	printEpochsat (&epoch->t, measType, ns, sats, options);
}




/*****************************************************************************
 * Name        : designSystemIniValues
 * Description : Updates the filter with the information of the former epoch 
 *               (initial values for the current one)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * double  *atwa                   IO N/A  At*W*A matrix
 * double  *atwy                   IO N/A  At*W*Y vector
 * TUnkinfo  *unkinfo              I  N/A  Structure containing the information of
 *                                         the unknowns 
 * TFilterSolution  *solution      I  N/A  Structure with the information of the 
 *                                         solution of the previous epoch
 *****************************************************************************/
void designSystemIniValues (TEpoch *epoch, double *atwa, double *atwy, TUnkinfo *unkinfo, TFilterSolution *solution) {
	int			i,j;
	double		corr[MAX_UNK];

	for (i=0;i<unkinfo->nunk;i++) {
		for (j=0;j<unkinfo->nunk;j++) {
			corr[j]=solution->correlations[m2v(i,j)];
		}
		atwa_atwy_insertLineWithCorrelation(atwa,atwy,corr,solution->parameters[i],unkinfo->nunk,1,i);
	}
}

/*****************************************************************************
 * Name        : calculatePostfits
 * Description : Calculate and print the postfits after the filtering process
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * double  *solution               I  N/A  Solution of the filter
 * TUnkinfo  *unkinfo              I  N/A  Structure containing the information of
 *                                         the unknowns
 * MeasurementType  measType       I  N/A  Measurement type of the postfit
 * int  filterInd                  I  N/A  Index of the measurement of the filter
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (double)         O  m    Standard deviation of the postfits
 *****************************************************************************/
double calculatePostfits(TEpoch *epoch, double *solution, TUnkinfo *unkinfo, enum MeasurementType measType, int filterInd, TOptions *options) {
	double		stddev2postfit=0;
 	double		postfit;
 	double		measurement,model;
	double		estimatedParameters;
	double		geometricCorrection;
	double		smoothWith;
	TSatellite	*sat;
	int			i;
	int			k;
	int			ret;

	for (i=0;i<epoch->numSatellites;i++) {
		sat = &epoch->sat[i];
		if (sat->available) {
			estimatedParameters = solution[unkinfo->par2unk[DT_UNK]];
			geometricCorrection = 0;
			for (k=0;k<3;k++) {
				geometricCorrection += -sat->LoS[k]*solution[unkinfo->par2unk[DR_UNK]+k];
			}
			estimatedParameters += geometricCorrection;
			if(options->estimateTroposphere) {
				estimatedParameters += sat->tropWetMap * solution[unkinfo->par2unk[TROP_UNK]];
			}
			if(whatIs(measType)==CarrierPhase) {
				estimatedParameters += solution[unkinfo->par2unk[BIAS_UNK]+i];
			}
			
			getMeasModelValue(epoch,sat->GNSS,sat->PRN,measType,&measurement,&model);
			
			// Smoothing
			if (options->smoothEpochs!=0 && options->smoothMeas[filterInd]!=NA) {
				ret = getMeasModelValue(epoch,sat->GNSS,sat->PRN,options->smoothMeas[filterInd],&smoothWith,NULL);
				if (ret!=1) {
					sat->available=0;
					model = 0;
					smoothWith = 0;
					sat->smoothedMeas[filterInd] = 0;
				}
				postfit = sat->smoothedMeas[filterInd] + smoothWith - (model + estimatedParameters);
			} else {
				postfit = measurement - (model + estimatedParameters);
			}
			
			if (whatIs(measType)==CarrierPhase) {
				printPostfit (&epoch->t, sat, measType, postfit, measurement, model, estimatedParameters, solution[unkinfo->par2unk[BIAS_UNK]+i], options);
			} else {
				printPostfit (&epoch->t, sat, measType, postfit, measurement, model, estimatedParameters, 0, options);
			}

			stddev2postfit += postfit*postfit;
		}
	}
	stddev2postfit /= epoch->usableSatellites;
	return stddev2postfit;
}

/*****************************************************************************
 * Name        : computeSolution
 * Description : Obtain the filter solution for the specified epoch
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * double  *stddev2postfit         O  m    Standard deviation of the postfits
 * TFilterSolution  *solution      IO N/A  Structure with the information of the 
 *                                         solution of the previous epoch (it is 
 *                                         updated with the solution of the present epoch)
 * double  *newcorrelations        O  N/A  Correlations after the filter run
 * double  *newparameters          O  N/A  Parameters after the filter run
 * TUnkinfo  *unkinfo              I  N/A  Structure containing the information of
 *                                         the unknowns
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (int)            O  N/A  Status of the function
 *                                         0 => Problem in the matrix inversion
 *                                         1 => Solution found
 *****************************************************************************/
int computeSolution (TEpoch *epoch,double *stddev2postfit, TFilterSolution *solution, double *newcorrelations, double *newparameterValues, TUnkinfo *unkinfo, TOptions *options) {
	double *atwa;
	double *atwy;
	double *aux;
	int 	i,j;
	int		res;

	atwy=malloc(sizeof(double)*unkinfo->nunk);
	atwa=malloc(sizeof(double)*unkinfo->nunkvector);

	// Initializing At*W*A and At*W*Y
	for (i=0;i<unkinfo->nunk;i++) {
		atwy[i]=0;
	}
	for (i=0;i<unkinfo->nunkvector;i++) {
		atwa[i]=0;
	}
	
	// Forming At*W*A and At*W*Y
	for (i=0;i<options->totalFilterMeasurements;i++) {
		designSystem(epoch,atwa,atwy,unkinfo,options->measurement[i],i,options);
	}

	designSystemIniValues(epoch,atwa,atwy,unkinfo,solution);

	res = cholinv_opt(atwa,unkinfo->nunk);

	if (res==-1) {
		free(atwa);
		free(atwy);
		sprintf(messagestr,"Problem in CHOLESKI1 res=%d\n",res);
		printError(messagestr,options);
		return 0;
	}
	
	// Saving parameters formal errors
//	for (i=0;i<unkinfo->nunk;i++) {
//		solution->formalError[i] = sqrt(atwa[m2v(i,i)]);
//	}
	
	mxv(atwa,atwy,unkinfo->nunk);

	// Updating correlations
	if (solution->correlations!=NULL) {
		for (i=0;i<unkinfo->nunk;i++) {
			for (j=0;j<=i;j++) {
				newcorrelations[m2v(i,j)] = atwa[m2v(i,j)];
			}
		}
		
		memcpy (newparameterValues,atwy,sizeof(double)*unkinfo->nunk);
	}

	// Building solution vector
	for (i=0;i<unkinfo->nunk;i++) {
		solution->x[i] = atwy[i];
	}
	for (i=0;i<3;i++) {
		solution->x[unkinfo->par2unk[DR_UNK]+i] = epoch->receiver.aproxPosition[i] + atwy[unkinfo->par2unk[DR_UNK]+i];
	}
	if (options->estimateTroposphere) solution->x[unkinfo->par2unk[TROP_UNK]] += epoch->ZTD;

	// High 'dr' checking => Initial aproximate position far from estimated position
	aux = &atwy[unkinfo->par2unk[DR_UNK]];
//	if (aux[0]*aux[0]+aux[1]*aux[1]+aux[2]*aux[2] > options->maxdr2) {
//		for (i=0;i<3;i++) epoch->receiver.aproxPosition[i] = solution->x[unkinfo->par2unk[DR_UNK]+i];
//			printf("MAXDR FOUND: %lf>%lf\n",aux[0]*aux[0]+aux[1]*aux[1]+aux[2]*aux[2],options->maxdr2);
//		computeSolution (epoch,stddev2postfit,solution,newcorrelations,newparameterValues,unkinfo,options);
//	}

	// Postfit calculation
	for (i=0;i<options->totalFilterMeasurements;i++) {
		*stddev2postfit = calculatePostfits(epoch,atwy,unkinfo,options->measurement[i],i,options);
	}
	// Formal Error Calculation
	solution->dop=0;

	for (i=0;i<3;i++) {
		solution->dop += atwa[m2v(unkinfo->par2unk[DR_UNK]+i,unkinfo->par2unk[DR_UNK]+i)];
	}
	solution->dop = sqrt(solution->dop);

	// Horizontal and Vertical Dilution Of Precision Computations
	calculateDOP(epoch,solution);

	
	// Freeing memory
	free(atwa);
	free(atwy);
	
	return 1;
}

/*****************************************************************************
 * Name        : calculateDOP
 * Description : Computes the Dilution of Precision of the current epoch  
 *               obtain the solution (only for the used satellites)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TFilterSolution  *solution      I  N/A  Structure with the information of the 
 *                                         solution of the previous epoch (it is 
 *                                         updated with the solution of the present epoch)
 *****************************************************************************/
void calculateDOP (TEpoch *epoch, TFilterSolution *solution) {
	double                sum;
	int		      i,j,k,vectorind;
        double        	      RowMatrixG_NEU[4], RowMatrixG_XYZ[4];
        double        	      VectorGtG_NEU[10], VectorGtG_XYZ[10], FalsePrefit[4];	

	//Initialize Variables
	solution->GDOP = 0.0; solution->PDOP = 0.0; solution->TDOP = 0.0;
	solution->HDOP = 0.0; solution->VDOP = 0.0;

	for (j=0;j<4;j++)   { RowMatrixG_NEU[j] = 0.0;   RowMatrixG_XYZ[j] = 0.0; FalsePrefit[j] = 0.0; }
	for (j=0;j<10;j++)  {  VectorGtG_NEU[j] = 0.0;   VectorGtG_XYZ[j] = 0.0; }

	// Build Geometry Matrix and its Transpose, then store its multiplication in a vector
	for (i=0;i<epoch->numSatellites;i++) 
		if (epoch->sat[i].available) {
			RowMatrixG_NEU[0] = cos ( epoch->sat[i].elevation ) * sin ( epoch->sat[i].azimuth ) ;
			RowMatrixG_NEU[1] = cos ( epoch->sat[i].elevation ) * cos ( epoch->sat[i].azimuth ) ;
			RowMatrixG_NEU[2] = sin ( epoch->sat[i].elevation ) ;
			RowMatrixG_NEU[3] = 1 ;

			RowMatrixG_XYZ[0] = epoch->sat[i].LoS[0] ;
			RowMatrixG_XYZ[1] = epoch->sat[i].LoS[1] ;
			RowMatrixG_XYZ[2] = epoch->sat[i].LoS[2] ;
			RowMatrixG_XYZ[3] = 1 ;

			atwa_atwy_insertline (VectorGtG_XYZ, FalsePrefit, RowMatrixG_XYZ, 0.0, 4, 1.0);
			atwa_atwy_insertline (VectorGtG_NEU, FalsePrefit, RowMatrixG_NEU, 0.0, 4, 1.0);
		}

	// Invert the Gt*G Matrixes:
	cholinv_opt(VectorGtG_NEU,4);
	cholinv_opt(VectorGtG_XYZ,4);

	// Compute DOPs
	solution->HDOP = sqrt ( VectorGtG_NEU[0] + VectorGtG_NEU[2] );
	solution->VDOP = sqrt ( VectorGtG_NEU[5] );
	solution->GDOP = sqrt ( VectorGtG_XYZ[0] + VectorGtG_XYZ[2] + VectorGtG_XYZ[5] + VectorGtG_XYZ[9] );
	solution->PDOP = sqrt ( VectorGtG_XYZ[0] + VectorGtG_XYZ[2] + VectorGtG_XYZ[5] );
	solution->TDOP = sqrt ( VectorGtG_XYZ[9] );

}


/*****************************************************************************
 * Name        : updatePRNlist
 * Description : Update the PRNlist vector for the list of satellites visible 
 *               in the current epoch
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * int  *PRNlist                   O  N/A  List of satellites
 * TFilterSolution  *solution      I  N/A  Structure with the information of the 
 *                                         solution
 *****************************************************************************/
void updatePRNlist(TEpoch *epoch, int *PRNlist, TFilterSolution *solution) {
	int i;

	solution->prevNumSatellites = 0;
	for (i=0;i<epoch->numSatellites;i++) {
		if (epoch->sat[i].GNSS == GPS) {
			solution->prevNumSatellites++;
			PRNlist[i]=epoch->sat[i].PRN;
		} else {
			PRNlist[i]=0;
		}
	}
}

/*****************************************************************************
 * Name        : Kalman
 * Description : Process the filter for the specified epoch
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TFilterSolution  *solution      IO N/A  Structure with the information of the 
 *                                         previous solution (updated)
 * int  *PRNlist                   IO N/A  List of satellites of the previous epoch
 *                                         (updated)
 * TUnkinfo  *prevUnkinfo          IO N/A  Structure containing the information
 *                                         of the unknowns of the previous epoch
 *                                         (updated)
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (int)            O  N/A  Status of the function
 *                                         0 => Epoch not computable
 *                                         1 => Epoch computed
 *                                         2 => Initial position wrong, model and Kalman must be relaunched
 *****************************************************************************/
int Kalman (TEpoch *epoch, TFilterSolution *solution, int *PRNlist, TUnkinfo *prevUnkinfo, TOptions *options) {
	double			newcorrelations[MAX_VECTOR_UNK];
	double			stddev2postfit;
	double			newparameterValues[MAX_UNK];
	TUnkinfo		unkinfo;
	TFilterSolution	aux;
	TEpoch			auxEpoch;
	static int		first = 1;
	static TTime	prevT;
	double 			distance;
	int				i;
	
	if (first) {
		memcpy(&prevT,&epoch->t,sizeof(TTime)); 
	}
	if (options->receiverPositionSource == rpCALCULATE) {
		memcpy(&auxEpoch,epoch,sizeof(TEpoch));
	}
	
	fillUsableSatellites(epoch,options);
	if (isEpochComputable(epoch)) {
		memcpy(&aux,solution,sizeof(TFilterSolution));	
		initUnkinfo(epoch,&unkinfo,calculateUnknowns(epoch->numSatellites,options),epoch->numSatellites,fabs(tdiff(&epoch->t,&prevT)),options);
		prepareCorrelation(epoch,solution,PRNlist,&unkinfo,prevUnkinfo,options);
		if (!computeSolution(epoch,&stddev2postfit,solution,newcorrelations,newparameterValues,&unkinfo,options)) {
			memcpy(solution,&aux,sizeof(TFilterSolution));
			return 0;
		}
		
		if (options->receiverPositionSource == rpCALCULATE) {
			distance = sqrt((solution->x[0]-epoch->receiver.aproxPosition[0])*(solution->x[0]-epoch->receiver.aproxPosition[0]) + 
							(solution->x[1]-epoch->receiver.aproxPosition[1])*(solution->x[1]-epoch->receiver.aproxPosition[1]) + 
							(solution->x[2]-epoch->receiver.aproxPosition[2])*(solution->x[2]-epoch->receiver.aproxPosition[2]));
			for (i=0;i<3;i++) {
				epoch->receiver.aproxPosition[i] = solution->x[i];
			}
			epoch->receiver.aproxPositionError = solution->dop;
			sprintf(messagestr,"Apriori position updated to: %13.4lf %13.4lf %13.4lf   Distance: %13.4lf",epoch->receiver.aproxPosition[0],epoch->receiver.aproxPosition[1],epoch->receiver.aproxPosition[2],distance);
			printInfo(messagestr,options);
			if (distance > 100) {
				memcpy(auxEpoch.receiver.aproxPosition,epoch->receiver.aproxPosition,sizeof(double)*3);
				auxEpoch.receiver.aproxPositionError = epoch->receiver.aproxPositionError;
				memcpy(epoch,&auxEpoch,sizeof(TEpoch));
				memcpy(solution,&aux,sizeof(TFilterSolution));
				return 2;
			}
		}

		printFilter(&epoch->t,unkinfo.nunk,solution,options);

		// Preparing values for next epoch
		memcpy(prevUnkinfo,&unkinfo,sizeof(TUnkinfo));
		memcpy(solution->correlations,newcorrelations,sizeof(double)*unkinfo.nunkvector);
		memcpy(solution->parameters,newparameterValues,sizeof(double)*unkinfo.nunk);

		updatePRNlist(epoch,PRNlist,solution);
		
		// Clean first flag
		first = 0;
		
		// Save last computed time
		memcpy(&prevT,&epoch->t,sizeof(TTime)); 
		return 1;
	} else {
		return 0;
	}
}
