/****************************************************************************
   Copyright & License:
   ====================
   
   Copyright 2009 - 2015 gAGE/UPC & ESA
   
   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 ***************************************************************************/

/****************************************************************************
 * Copyright: gAGE/UPC & ESA
 * Project: EDUNAV GNSS Lab Tool 
 * Author: Pere Ramos-Bosch (group of Astronomy and GEomatics - gAGE/UPC)
 * Curator: Adria Rovira-Garcia ( gAGE/UPC )
 * Developer: Deimos Ibanez Segura ( gAGE/UPC )      
 *          glab.gage @ upc.edu
 * File: preprocessing.c
 * Code Management Tool File Version: 2.2  Revision: 3
 * Date: 2015/05/28
 ***************************************************************************/

/****************************************************************************
 * MODULE DESCRIPTION
 *
 * Name: preprocessing
 * Language: C
 *
 * Purpose:
 *  The purpose of this module is to preprocess the input data
 *
 * Function:
 *  This class is in charge of preprocessing the input data by doing several
 *  steps:
 *     - Cycle-slip detection
 *     - Inconsistent pseudorange-carrierphase jump detection (for Septentrio receivers)
 *     - Epoch decimation
 *     - Carrier-phase prealignment
 *     - Pseudorange smoothing
 *
 * Dependencies:
 *  dataHandling
 *
 * Files modified:
 *  None
 *
 * Files read:
 *  Depending on function
 *
 * Resources usage:
 *  See Design Document
 ****************************************************************************/

/****************************
 *       RELEASE_HISTORY
 * -------------
 *          gLAB v1.3.1
 * Release: 2010/01/22
 * Change Log: First public version.
 * -------------
 *          gLAB v1.3.2
 * Release: 2010/02/23
 * Change Log: Changed the time to start of the Li and L1-C1 from the 3rd epoch
 *                of the arch to the forth.
 * -------------
 *          gLAB v1.4.0
 * Release: 2010/06/21
 * Change Log: Included prealignCP option to set/unset carrier-phase prealignment.
 *             Cycle-slip detector changed to specify when it does not have enough measurements.
 * -------------
 *          gLAB v1.4.1
 * Release: 2010/07/09
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.2
 * Release: 2010/07/31
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.3
 * Release: 2010/08/31
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.4
 * Release: 2010/09/22
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v1.4.5
 * Release: 2010/10/31
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.0.0
 * Release: 2012/12/31
 * Change Log: gLAB version released together with the Educational Book:
 *             "GNSS Data Processing" (Vol. 1 and Vol. 2). ESA TM-23.
 *             Authors: J. Sanz Subirana, J.M. Juan Zornoza and M. Hernandez-Pajares
 *             ISBN: 978-92-9221-885-0 (two volumes)
 *             ISSN: 1013-7076
 *             December 2012.
 * -----------
 *          gLAB v2.2.0
 * Release: 2014/09/22
 * Change Log: Modified checkCycleSlips function in order to weight current sigma with an initial value.
 * -----------
 *          gLAB v2.2.1
 * Release: 2014/12/23
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.2
 * Release: 2015/03/02
 * Change Log: No changes in this file.
 * -----------
 *       END_RELEASE_HISTORY
 *****************************/

/* External classes */
#include "preprocessing.h"

/**************************************
 * Declarations of internal operations
 **************************************/
 
/*****************************************************************************
 * Name        : preprocess
 * Description : Batch that runs a set of general preprocess functions 
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch reference structure
 * TOptions  *options              I  N/A  TOptions structure for preprocessing
 *                                         configuration
 *****************************************************************************/
void preprocess (TEpoch *epoch, TOptions *options) {
	if (options->prealignCP) prealignEpoch(epoch);
	checkPseudorangeJumps (epoch,options);
	checkCycleSlips(epoch,options);
	smoothEpoch(epoch,options);
}

/*****************************************************************************
 * Name        : isEpochDecimated
 * Description : Returns if the current is decimated or not depending on the
 *               chosen value of the options structure 
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch reference structure
 * TOptions  *options              I  N/A  TOptions structure for configuration
 * Returned value (int)            O  N/A  Status of the function
 *                                         0 => Epoch valid
 *                                         1 => Epoch to decimate  
 *****************************************************************************/
int isEpochDecimated (TEpoch *epoch, TOptions *options) {
	int		nearestInt;
	
	if (options->decimate==0) return 0;
	else {
		nearestInt = (int)(epoch->t.SoD+0.5);
		if ((nearestInt%options->decimate)==0) return 0;
		else return 1;
	}
}

/*****************************************************************************
 * Name        : isEclipsed
 * Description : Check if a satellite is under eclipse (or has been for the
 *               last 30 minutes)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  Reference time
 * TSatellite  *sat                I  N/A  Reference satellite
 * TSatInfo  *satInfo              I  N/A  Information of last eclipse
 * double  sunPos[3]               I  km   Sun position
 * double  *tLast                  O  s    Time from last eclipse (0 if it is
 *                                         currently in eclipse). Unmodified
 *                                         if no eclipse
 * Returned value (int)            O  N/A  Status of the function
 *                                    0 => Satellite OK
 *                                    1 => Satellite under eclipse
 *****************************************************************************/
int isEclipsed(TTime *t, TSatellite *sat, TSatInfo *satInfo, double sunPos[3], double *tLast) {
	double	satPosU[3];
	double	modSat;
	double	sunPosU[3];
	double	cosT;
	double	val;
	double  diff;

	if (!sat->hasOrbitsAndClocks) return 0; // Unable to know if it is eclipsed

	// Check if it is now in eclipse
	modSat = sqrt(sat->position[0]*sat->position[0] + sat->position[1]*sat->position[1] + sat->position[2]*sat->position[2]);
	
	memcpy(satPosU,sat->position,sizeof(double)*3);
	memcpy(sunPosU,sunPos,sizeof(double)*3);
	
	unitary (satPosU);
	unitary (sunPosU);
	cosT = scalarProd(satPosU,sunPosU);
	
	if (cosT<0) {
		val = modSat*sqrt(1-cosT*cosT);
		if (val < EARTH_RADIUS) {
			memcpy(&satInfo->lastEclipse,t,sizeof(TTime));
			*tLast = 0;
			return 1;
		}
	}
	
	// Check if it has been in eclipse in the last 30 minutes
	if (satInfo->lastEclipse.MJDN!=0) { // This would mean still no eclipse found for this satellite
		diff = fabs(tdiff(t,&satInfo->lastEclipse));
		if (diff <= 1800) { // 30 minutes
			*tLast = diff;
			return 1;
		} 
	}
	
	return 0;
}

/*****************************************************************************
 * Name        : prealignSat
 * Description : Prealign pseudorange and carrierphase from a single satellite
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch structure
 * int  ind                        I  N/A  Index of the satellite
 *****************************************************************************/
void prealignSat (TEpoch *epoch, int ind) {
	int						j;
	enum GNSSystem			GNSS;
	enum MeasurementType	meas;
	double					CP,PR;
	double					lambda;
	int						ret1,ret2;
	
	if (epoch->prealign==0) return;
	if (epoch->sat[ind].arcLength == 1) {
		GNSS = epoch->sat[ind].GNSS;
		for (j=0;j<epoch->measOrder[GNSS].nDiffMeasurements;j++) {
			meas = epoch->measOrder[GNSS].ind2Meas[j];
			//if (epoch->measOrder[GNSS].usable[meas]==1) { // Measurement has NOT been invalidated
				if (whatIs(meas)==CarrierPhase) {
					// Set prealign value to 0
					epoch->sat[ind].phasePrealign[j] = 0;
					ret1 = getMeasModelValue(epoch,GNSS,epoch->sat[ind].PRN,meas,&CP,NULL);
					ret2 = getMeasModelValue(epoch,GNSS,epoch->sat[ind].PRN,meas-1,&PR,NULL);
					// Special cases: 
					// For the case of L1P, for GPS, if it cannot prealign with C1P (P1 code), it
					// will try to prealign it with C1C (C1 code).
					if (meas==L1P && ret2==0) {
						ret2 = getMeasModelValue(epoch,GNSS,epoch->sat[ind].PRN,C1C,&PR,NULL);
					}
					// For the case of L2P, for GPS, if it cannot prealign with C2P (P2 code), it
					// will try to prealign it with C2C (C2 code).
					if (meas==L2P && ret2==0) {
						ret2 = getMeasModelValue(epoch,GNSS,epoch->sat[ind].PRN,C2C,&PR,NULL);
					}
					// Obtain prealign value
					if (ret1!=0 && ret2!=0) { 
						lambda = getLambda(GNSS,meas);
						epoch->sat[ind].phasePrealign[j] = ((int)((PR-CP)/lambda));
			//		} else { // Mark CP as invalid if it was not
			//			epoch->sat[ind].meas[j].value = -1;
			//			initSatellite(epoch,ind);
			//			epoch->sat[ind].arcLength = 0; // This is to invalidate current epoch
					}
				}		
			//}
		}
	}
}

/*****************************************************************************
 * Name        : prealignEpoch
 * Description : Prealign pseudorange and carrierphase an entire epoch
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch structure
 *****************************************************************************/
void prealignEpoch (TEpoch *epoch) {
	int						i;
	
	epoch->prealign = 1;
	for (i=0;i<epoch->numSatellites;i++){
		prealignSat(epoch,i);
	}
}

/*****************************************************************************
 * Name        : lagrangeInterpolation
 * Description : Performs a Lagrange interpolation of specified degree
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int  degree                     I  N/A  Defree of the interpolation
 * TTime  t                        I  N/A  Target x point to interpolate in TTime
 *                                         format
 * TTime  *tPrev                   I  N/A  x axis in TTime format
 * double  *yPrev                  I  N/A  y axis
 * Returned value (double)         O  N/A  Interpolated value at t
 *****************************************************************************/
double lagrangeInterpolation (int degree, TTime t, TTime *tPrev, double *yPrev) {
	int			i,j;
	double		yEst,prod;
	double		*xPrev,x;
	TTime		tRef;
	
	yEst = 0;
	xPrev = malloc(sizeof(double)*(degree+1));
	
	// Fill xPrev and x with tPrev and t values   
	memcpy(&tRef,&tPrev[0],sizeof(TTime)); // Reference time
	x = tdiff(&t,&tRef);
	for (i=0;i<=degree;i++)
		xPrev[i] = tdiff(&tPrev[i],&tRef);
				
	// Perform the interpolation
	for (i=0;i<=degree;i++) {
		prod = 1;
		for (j=0;j<=degree;j++) {
			if (i!=j) {
				prod *= (x-xPrev[j])/(xPrev[i]-xPrev[j]); 
			}
		}
		prod *= yPrev[i];
		yEst += prod;
	}

	free(xPrev);
	
	return yEst;
}

/*****************************************************************************
 * Name        : checkPseudorangeJumps
 * Description : Check for carrier-phase / pseudorange inconsistencies
 *               Some receivers do not output a set of physically consistent 
 *               measurements, when they adjust their clock a "jump" should 
 *               appear both in pseudorange and carrier-phase measurements,
 *               but it only appears in pseudorange.
 *               This creates a massive cycle-slip for all satellites in this 
 *               epoch if it is not corrected.
 *               This function detects and corrects this behaviour, providing a 
 *               set of consistent observables.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch structure
 * TOptions  *options              I  N/A  TOptions structure 
 * Returned value (int)            O  N/A  Status of the function
 *                                         0 => No pseudorange jump found
 *                                         1 => Pseudorange jump found and corrected
 *****************************************************************************/
int checkPseudorangeJumps (TEpoch *epoch, TOptions *options) {
	static const double		sigmaLimit2 = 100*100;
	int						i,k;
	enum MeasurementType	type;
	enum MeasurementKind	kind;
	int						numPR;
	int						numCP;
	int						numTot = 0;
	double					meanPR = 0;
	double					meanCP = 0;
	double					meanTot = 0;
	double					mean2Tot = 0;
	double					sigma2Tot;
	double					measurement;
	int						res;
	int						corrected = 0;
	double					correction = 0;
	double					lambda;
	
	for (i=0;i<epoch->numSatellites;i++) {
		numPR = 0;
		numCP = 0;
		if (epoch->sat[i].GNSS == GPS && epoch->sat[i].arcLength >= 3) { // Only GPS
			for (k=0;k<epoch->measOrder[epoch->sat[i].GNSS].nDiffMeasurements;k++) {
				type = epoch->measOrder[epoch->sat[i].GNSS].ind2Meas[k];
				kind = whatIs(type);
				res = getMeasModelValue(epoch,epoch->sat[i].GNSS,epoch->sat[i].PRN,type,&measurement,NULL);
				if (res==1) {
					if (kind==Pseudorange) {
						meanPR = (numPR*meanPR + measurement)/(numPR+1);
						numPR++;
					} else if (kind==CarrierPhase) {
						meanCP = (numCP*meanCP + measurement)/(numCP+1);
						numCP++;
					}
				}
			}
			if (numCP>0 && numPR>0 && epoch->sat[i].arcLength!=1) {
				meanTot = (numTot*meanTot + (meanPR-meanCP))/(numTot+1);
				mean2Tot = (numTot*mean2Tot + (meanPR-meanCP)*(meanPR-meanCP))/(numTot+1);
				numTot++;
			}
		}
	}
	
	sigma2Tot = mean2Tot-meanTot*meanTot;
	if (sigma2Tot<sigmaLimit2 && numTot>1) {
		correction = (int)(meanTot / (c*1e-3)+0.5);
		if (correction!=0) {
			corrected = 1;
			sprintf(messagestr,"Inconsistence in carrier-phase pseudorange detected at %lf. The possible clock adjustement was corrected (%lf ms)",epoch->t.SoD,correction);
			printInfo(messagestr,options);
			correction = correction * (c * 1e-3);
			// Correct observables from jump only in pseudorange
			for (i=0;i<epoch->numSatellites;i++) {
				if (epoch->sat[i].GNSS == GPS) { // Only GPS
					for (k=0;k<epoch->measOrder[epoch->sat[i].GNSS].nDiffMeasurements;k++) {
						type = epoch->measOrder[epoch->sat[i].GNSS].ind2Meas[k];
						kind = whatIs(type);
						if (kind==CarrierPhase) {
							lambda = getLambda(epoch->sat[i].GNSS,type);
							epoch->sat[i].phasePrealign[k] += correction/lambda;
						}
					}
				}
			}
		}
	}
	return corrected;
}

/*****************************************************************************
 * Name        : checkCycleSlips
 * Description : Check for carrier-phase cycle slips. At the moment it is only 
 *               prepared for GPS
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TOptions  *options              I  N/A  TOptions structure 
 * Returned value (int)            O  N/A  Status of the function
 *                                         0 => No cycle-slips in current epoch
 *                                         1 => Cycle slip found in any measurement
 *****************************************************************************/
int checkCycleSlips(TEpoch *epoch, TOptions *options) {
	int				i;
	enum GNSSystem	system;
	int				cycleSlipFound = 0;
	double			dt;
	
	// BW
	double			Bw;
	double			nBw;
	double			nBwdiff;
	double			nBwdiffThreshold;
	int				BwCheck;

	// Li
	double			Li, LiEst;
	double			LiThreshold;
	int				LiCheck;
	int				liDegree;
	
	// L1-C1
	double			L1, C1;
	double			L1C1diff;
	double			L1C1Threshold;
	int				L1C1Check;
	int				nL1C1;
	
	int				effArcLength;
	double			sigma2;
	
	// Go over all the satellites
	for (i=0;i<epoch->numSatellites;i++) {
		system = epoch->sat[i].GNSS;
		// Check initialization
		LiCheck = 0;
		BwCheck = 0;
		L1C1Check = 0;
		if (system == GPS) {
			// Melbourne-Wübbena (BW) [dual-frequency]
			if (options->csBW) {
				Bw = getMeasurementValue(epoch,system,epoch->sat[i].PRN,BW);
				if (Bw==-1) {
					BwCheck = 1;
					nBwdiffThreshold = -1; // The theshold is set to -1 to indicate that there are not enough measurements available
				} else {
					nBw = Bw / GPSlw;
					if (epoch->sat[i].arcLength == 1) {
						epoch->cycleslip.nBwmean[i] = nBw;
						epoch->cycleslip.nBwmean2[i] = nBw*nBw;
					}
					nBwdiff = nBw - epoch->cycleslip.nBwmean[i];
					sigma2 = epoch->cycleslip.nBwmean2[i] - epoch->cycleslip.nBwmean[i]*epoch->cycleslip.nBwmean[i];
					nBwdiffThreshold = min(options->csBWmax,max(options->csBWslope*sqrt(sigma2),options->csBWmin));
					if (epoch->sat[i].arcLength >= 4) {
						if (fabs(nBwdiff)>nBwdiffThreshold) BwCheck = 1;
					}
				}
			}
			
			// Geometric-free carrier-phase combination (Li) [dual-frequency]
			liDegree = 2;
			if (options->csLi) {
				Li = getMeasurementValue(epoch,system,epoch->sat[i].PRN,LI);
				if (Li==-1) {
					LiCheck = 1;
					LiThreshold = -1; // The theshold is set to -1 to indicate that there are not enough measurements available
				} else {
					if (epoch->sat[i].arcLength >= 4) {
						if (tdiff(&epoch->cycleslip.tPrev[0],&epoch->cycleslip.tPrev[1])==0) liDegree=1;
						if (tdiff(&epoch->cycleslip.tPrev[1],&epoch->cycleslip.tPrev[2])==0) liDegree=1;
						if (tdiff(&epoch->cycleslip.tPrev[0],&epoch->cycleslip.tPrev[2])==0) liDegree=1;
						LiEst = lagrangeInterpolation(liDegree,epoch->t,epoch->cycleslip.tPrev,epoch->cycleslip.LiPrev[i]);
						dt = fabs(tdiff(&epoch->t,&epoch->cycleslip.tPrev[2]));
						LiThreshold = options->csLimax+(options->csLimin-options->csLimax)*exp(-dt/options->csLit);
						if (fabs(Li-LiEst)>LiThreshold) LiCheck = 1;
					}
				}
			}
			
			// L1-C1 Cycle-slip detector [single-frequency]
			if (options->csL1C1) {
				L1 = getMeasurementValue(epoch,system,epoch->sat[i].PRN,L1P);
				//L1 = getMeasurementValue(epoch,system,epoch->sat[i].PRN,L1C);
				C1 = getMeasurementValue(epoch,system,epoch->sat[i].PRN,C1C);
				if (L1==-1 || C1==-1) {
					L1C1Check = 1;
					L1C1Threshold = -1; // The theshold is set to -1 to indicate that there are not enough measurements available
				} else {
					if (epoch->sat[i].arcLength == 1) {
						epoch->cycleslip.L1C1mean[i] = L1-C1;
						epoch->cycleslip.L1C1mean2[i] = (L1-C1)*(L1-C1);
					}
					L1C1diff = (L1-C1) - epoch->cycleslip.L1C1mean[i];
					sigma2 = epoch->cycleslip.L1C1mean2[i] - epoch->cycleslip.L1C1mean[i]*epoch->cycleslip.L1C1mean[i];

                                        // Weight current sigma with an initial value
			                nL1C1 = epoch->sat[i].arcLength;
                                        sigma2 = ( (nL1C1-1)*sigma2 + options->csL1C1init*options->csL1C1init )/nL1C1;


					L1C1Threshold = min(options->csL1C1slope * sqrt(sigma2),options->csL1C1max);
					if (epoch->sat[i].arcLength >= 4) {
						if (fabs(L1C1diff)>L1C1Threshold) L1C1Check = 1; 
					}
				}
			}

			if (LiCheck || BwCheck || L1C1Check) {
				// Cycle-slip detected!
				initSatellite(epoch,i);
				prealignSat(epoch,i);
				epoch->cycleslip.LiPrev[i][MAX_CS_MEMORY-1] = getMeasurementValue(epoch,system,epoch->sat[i].PRN,LI);
				nBw = getMeasurementValue(epoch,system,epoch->sat[i].PRN,BW) / GPSlw;
				L1 = getMeasurementValue(epoch,system,epoch->sat[i].PRN,L1P);
				printCS(epoch,i,LiCheck,fabs(Li-LiEst),LiThreshold,BwCheck,fabs(nBwdiff),nBwdiffThreshold,L1C1Check,fabs(L1C1diff),L1C1Threshold,options);
				cycleSlipFound = 1;
			}

			// Update nBwmean and nBwsigma2
			if (options->csBW) {
				if (epoch->sat[i].arcLength >= 10) {
					effArcLength = 10;
				} else {
					effArcLength = epoch->sat[i].arcLength;
				}
				epoch->cycleslip.nBwmean[i] = ((effArcLength-1)*epoch->cycleslip.nBwmean[i] + nBw)/effArcLength;
				epoch->cycleslip.nBwmean2[i] = ((effArcLength-1)*epoch->cycleslip.nBwmean2[i] + nBw*nBw)/effArcLength;
			}
			
			// Update L1C1mean and L1C1mean2
			if (options->csL1C1) {
				if (epoch->sat[i].arcLength >= options->csL1C1window) {
					nL1C1 = options->csL1C1window; 
				} else {
					nL1C1 = epoch->sat[i].arcLength;
				}
				epoch->cycleslip.L1C1mean[i] = ((nL1C1-1)*epoch->cycleslip.L1C1mean[i] + (L1-C1)) / nL1C1;
				epoch->cycleslip.L1C1mean2[i] = ((nL1C1-1)*epoch->cycleslip.L1C1mean2[i] + (L1-C1)*(L1-C1)) / nL1C1;
			}
		} else { // System unsupported
		}
	}

	
	return cycleSlipFound;
}

/*****************************************************************************
 * Name        : smoothEpoch
 * Description : Smooth the pseudorange measurements to be used in the filter
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch structure
 * TOptions  *options              I  N/A  TOptions structure 
 *****************************************************************************/
void smoothEpoch (TEpoch *epoch, TOptions *options) {
	int			i,j;
	double		smoothWith;
	double		measurement;
	TSatellite	*sat;
	int			ret1,ret2;
	int			smoothN;
	
	for (i=0;i<options->totalFilterMeasurements;i++) {
		if (options->smoothEpochs!=0 && options->smoothMeas[i]!=NA) {
			for (j=0;j<epoch->numSatellites;j++) {
				sat = &epoch->sat[j];
				ret1 = getMeasModelValue(epoch,sat->GNSS,sat->PRN,options->smoothMeas[i],&smoothWith,NULL);
				ret2 = getMeasModelValue(epoch,sat->GNSS,sat->PRN,options->measurement[i],&measurement,NULL);
				if (ret1!=1 || ret2!=1) {
					sat->available=0;
				} else {
					smoothN = sat->arcLength;
					if (smoothN > options->smoothEpochs) {
						smoothN = options->smoothEpochs;
					}
					// Sanity check
					if (smoothN==0) smoothN=1;
					
					sat->smoothedMeas[i] = ((smoothN-1)*sat->smoothedMeas[i] + (measurement-smoothWith)) / (smoothN);
				}
			}
		}
	}
}
